#!/usr/bin/python
'''ubuntu-repository-cache charm

This 'hooks' source constitutes all of the code for handle charm hooks.
'''

import os
import pprint
import sys

sys.path.insert(0, os.path.join(os.environ['CHARM_DIR'], 'lib'))

# pylint can't find the modules  # pylint: disable=F0401
from charmhelpers.core import hookenv
from charmhelpers.contrib import unison
from ubuntu_repository_cache import (
    mirror,
    service,
    util,
)
# pylint: enable=F0401

HOOKS = hookenv.Hooks()
LOG = hookenv.log
SERVICE = 'ubuntu-repository-cache'
PRETTYPRINT = pprint.PrettyPrinter(indent=4)


@HOOKS.hook('debug')
def debug():
    '''Debug hook for development'''
    PRETTYPRINT.pprint(hookenv.execution_environment())


@HOOKS.hook('install')
def install():
    '''Perform service installation'''

    LOG('Installing %s' % SERVICE)
    service.install()


@HOOKS.hook('config-changed')
def config_changed():
    '''Handle charm configuration changes'''

    LOG('Config changed for %s' % SERVICE)
    service.render_configs()
    service.update_checks()


@HOOKS.hook('upgrade-charm')
def upgrade_charm():
    '''Handle charm updates'''

    LOG('Upgrading %s' % SERVICE)
    service.render_configs()
    service.update_checks()


@HOOKS.hook('ubuntu_repository_cache_sync')
@util.run_as_user('www-sync')
def ubuntu_repository_cache_sync():
    '''Sync Ubuntu repository metadata to the mirror.

    Metadata synchronization is performed and then the metadata is
    verified.'''

    LOG('Performing metadata sync for %s' % SERVICE)
    config = hookenv.config()
    apache_root = config['apache-root']
    apache_data = '/'.join((apache_root, 'data'))
    meta_ver = mirror.mirror_archive(hookenv.config()['sync-host'])
    if not meta_ver:
        LOG('Leader could not gather new metadata', hookenv.WARNING)
        return

    link_next = '/'.join((apache_data, 'ubuntu_next'))

    meta_dir = os.path.realpath(link_next)

    # Push the metadata to peers
    if hookenv.is_relation_made('cluster'):
        dest = '/'.join((apache_data, meta_ver))
        link_dir = '/'.join((apache_data, 'ubuntu_active'))
        LOG('Pushing updates to peers', hookenv.DEBUG)
        source = meta_dir + '/'
        successful_peers = mirror.rsync_to_peers(
            source, dest, link_dir=link_dir)

        # Push metaversion_good canary to indicate a complete sync
        canary = '_'.join((meta_dir, 'good'))
        LOG('Pushing canary files to peers', hookenv.DEBUG)
        mirror.rsync_to_peers(canary, apache_data, peers=successful_peers)

        # Update 'meta_version' relation setting
        LOG('', hookenv.DEBUG)
        ids = hookenv.relation_ids('cluster') or []
        LOG('Updating meta_version for ids %s' % ids, hookenv.DEBUG)
        for r_id in ids:
            LOG('Setting meta_version=%s' % meta_ver, hookenv.DEBUG)
            hookenv.relation_set(relation_id=r_id, meta_version=meta_ver)

    util.update_active_symlink(os.path.basename(meta_dir))
    util.clean_metadata_dir()
    service.start()


@HOOKS.hook('start')
def start():
    '''Start the charm's services'''

    LOG('Starting %s' % SERVICE)
    service.render_configs()

    sync_host = hookenv.config()['sync-host']
    config = hookenv.config()
    if config['sync-on-start']:
        LOG('Performing initial sync from %s' % sync_host)
        meta_ver = mirror.mirror_archive(sync_host)
        if meta_ver:
            LOG('Initial sync from {} completed ver={}.'.format(sync_host,
                                                                meta_ver))
            service.start()
        else:
            LOG('Leader could not gather new metadata', hookenv.WARNING)
    else:
        LOG('Skipping initial content sync, "sync-on-start" config '
            'option was not set')


@HOOKS.hook('stop')
def stop():
    '''Stop the charm's services'''

    LOG('Stopping %s' % SERVICE)
    service.stop()


@HOOKS.hook('website-relation-joined')
def website_relation_joined():
    '''Provide hostname/port for the website relationship'''

    LOG('Website relation joined for %s' % SERVICE)
    settings = {'hostname': hookenv.unit_private_ip(),
                'port': 80}
    hookenv.relation_set(relation_settings=settings)


@HOOKS.hook('cluster-relation-joined')
def cluster_relation_joined():
    '''Join a peer cluster relationship'''

    LOG('Cluster relation joined for %s' % SERVICE)
    unison.ssh_authorized_peers(peer_interface='cluster',
                                user='www-sync', group='www-data')
    mirror.peer_update_metadata()
    service.render_configs()


@HOOKS.hook('cluster-relation-changed')
def cluster_relation_changed():
    '''Handle peer cluster relationship changes'''

    LOG('Cluster relation changed for %s' % SERVICE)
    unison.ssh_authorized_peers(peer_interface='cluster',
                                user='www-sync', group='www-data')
    mirror.peer_update_metadata()
    service.render_configs()


@HOOKS.hook('cluster-relation-departed')
def cluster_relation_departed():
    '''Handle peer cluster relationship departures'''

    LOG('Cluster relation departed for %s' % SERVICE)
    unison.ssh_authorized_peers(peer_interface='cluster',
                                user='www-sync', group='www-data')
    mirror.peer_update_metadata()
    service.render_configs()


@HOOKS.hook('cluster-relation-broken')
def cluster_relation_broken():
    '''Leave a peer cluster relationship'''

    LOG('Cluster relation broken for %s' % SERVICE)
    service.render_configs()


@HOOKS.hook('nrpe-external-master-relation-changed')
@HOOKS.hook('local-monitors-relation-changed')
def update_checks():
    '''Update Nagios checks for all services'''

    service.update_checks()


if __name__ == "__main__":
    # execute a hook based on the name the program is called by
    HOOKS.execute(sys.argv)
