'''Functions related to apache use within this charm'''

import os
import subprocess

# pylint can't find the modules  # pylint: disable=F0401
from charmhelpers import (
    fetch,
)

from charmhelpers.core import (
    hookenv,
    host,
    templating
)
import util
# pylint: enable=F0401


LOG = hookenv.log
SERVICE = 'apache2'


@util.run_as_user('root')
def install():
    '''Perform the install/config of apache and disable the default site'''

    LOG('Installing apache', hookenv.INFO)
    fetch.apt_install(SERVICE, fatal=True)
    stop()
    subprocess.check_call(['a2enmod', 'expires'])
    subprocess.check_call(['a2enmod', 'proxy'])
    subprocess.check_call(['a2enmod', 'proxy_http'])
    subprocess.check_call(['a2enmod', 'headers'])
    subprocess.check_call(['a2dismod', 'mpm_event'])
    subprocess.check_call(['a2enmod', 'mpm_worker'])
    subprocess.check_call(['a2dissite', '000-default'])


@util.run_as_user('root')
def start():
    '''Start the service and remove the override file blocking startup'''

    LOG('Starting apache', hookenv.INFO)
    override_path = '/etc/init/apache2.override'
    if os.path.exists(override_path):
        os.remove('/etc/init/apache2.override')

    subprocess.call(['a2ensite', 'archive_ubuntu_com'])
    if not service_running():
        host.service_start(SERVICE)
    else:
        host.service_reload(SERVICE)


@util.run_as_user('root')
def stop():
    '''Stop the service and block startup at boot'''

    LOG('Stopping apache', hookenv.INFO)
    with open('/etc/init/apache2.override', 'w+') as override_file:
        override_file.write('manual\n')

    host.service_stop(SERVICE)
    subprocess.call(['a2dissite', 'archive_ubuntu_com'])


@util.run_as_user('root')
def service_running():
    '''Return True is the service is running'''

    return host.service_running(SERVICE)


def create_mpm_workerfile():
    '''Create the multi-processing module (MPM) configuration'''

    config = hookenv.config()
    mpm_context = {
        'startservers': config['apache2_mpm_startservers'],
        'minsparethreads': config['apache2_mpm_minsparethreads'],
        'maxsparethreads': config['apache2_mpm_maxsparethreads'],
        'threadlimit': config['apache2_mpm_threadlimit'],
        'threadsperchild': config['apache2_mpm_threadsperchild'],
        'serverlimit': config['apache2_mpm_serverlimit'],
        'maxrequestworkers': config['apache2_mpm_maxrequestworkers'],
        'maxconnectionsperchild': config['apache2_mpm_maxconnectionsperchild'],
    }
    templating.render('apache2/mpm_worker.template',
                      '/etc/apache2/conf-available/000mpm-worker.conf',
                      mpm_context)
    subprocess.check_call(['a2enconf', '000mpm-worker'])


def create_security():
    '''Create the security configuration'''

    config = hookenv.config()
    security_context = {
        'server_tokens': config['apache2_server_tokens'],
        'server_signature': config['apache2_server_signature'],
        'trace_enabled': config['apache2_trace_enabled'],
    }
    templating.render('apache2/security.template',
                      '/etc/apache2/conf-available/security.conf',
                      security_context)
    subprocess.check_call(['a2enconf', 'security'])


def create_metadata_site():
    '''Create the site config for serving (or proxying) repository metadata'''

    config = hookenv.config()
    apache_context = {}
    apache_context['DocumentRoot'] = config['apache-root']
    apache_context['Sync_Host'] = config['sync-host']
    metaproxy_conf = '# No failover configured'
    failover_host = util.get_failover_host()
    if util.in_failover() and failover_host:
        uri = 'http://%s:80' % failover_host
        metaproxy_conf = ('ProxyRemote      http://archive.ubuntu.com/ %s\n'
                          '\tProxyPass        / http://archive.ubuntu.com/\n'
                          '\tProxyPassReverse / http://archive.ubuntu.com/' %
                          uri)
    apache_context['MetadataProxy'] = metaproxy_conf

    templating.render('apache2/archive_ubuntu_com.conf',
                      '/etc/apache2/sites-available/archive_ubuntu_com.conf',
                      apache_context)


@util.run_as_user('root')
def render_configs():
    '''Render the configuration templates for apache'''

    LOG('Rendering apache2 configuration templates')
    config = hookenv.config()
    if config.changed('apache2_mpm_type'):
        if config['apache2_mpm_type'] == 'worker':
            LOG('Using apache2-mpm-worker')
            fetch.apt_install('apache2-mpm-worker', fatal=True)
            subprocess.check_call(['a2dismod', 'mpm_prefork'])
            subprocess.check_call(['a2enmod', 'mpm_worker'])
            fetch.apt_purge('apache2-mpm-prefork', fatal=True)
        elif config['apache2_mpm_type'] == 'prefork':
            LOG('Using apache2-mpm-prefork')
            fetch.apt_install('apache2-mpm-prefork', fatal=True)
            subprocess.check_call(['a2dismod', 'mpm_worker'])
            subprocess.check_call(['a2enmod', 'mpm_prefork'])
            fetch.apt_purge('apache2-mpm-worker', fatal=True)

    create_metadata_site()
    create_mpm_workerfile()
    create_security()


def update_checks(nrpe_config):
    '''Update nagios check for apache serving archive metadata'''

    nrpe_config.add_check(
        shortname='apache',
        description='Apache2 serving archive metadata',
        check_cmd='check_http --hostname localhost --url "/ubuntu/" '
                  '--string "dists/" --expect \"200\\ OK\"',
        )
