'''Functions relating to mirrored Ubuntu package repository content'''

import errno
import hashlib
import glob
import os
import socket
import sys
import time
import traceback

from collections import OrderedDict
from datetime import datetime
from subprocess import CalledProcessError

# pylint: disable=F0401
from charmhelpers.contrib import unison
from charmhelpers.contrib.hahelpers import cluster

from charmhelpers.core import (
    hookenv,
    host,
)

import service
import util
# pylint: enable=F0401

LOG = hookenv.log
SERVICE = 'ubuntu-repository-cache'


class ContentBad(StandardError):
    '''Mirror content is inconsistent'''
    pass


@util.run_as_user('www-sync')
def mirror_base(archive_ip, dest):
    '''Mirror just the root of the archive.

    This checks for the file 'Archive-Update-in-Progress-*' and
    throws ContentBad if the archive source isn't ready.
    Use this as a quick test prior to attempting a full mirror
    operation.
    '''

    source = archive_ip + '::ubuntu/'
    host.mkdir(dest, owner='www-sync', group='www-data', perms=0o775)
    apache_data = '/'.join((hookenv.config()['apache-root'], 'data'))
    link_dest = '/'.join((apache_data, 'ubuntu_active'))

    # Sync just the root directory
    rsync_options = ('--links', '--perms', '--times', '--dirs', '--chmod=g+w',
                     '--delete', '--compress', '--link-dest=%s/' % link_dest)
    host.rsync(source, dest, flags='--quiet', options=rsync_options)

    # Check for file 'Archive-Update-in-Progress*' which signals
    # the need to sleep or try another IP
    if glob.glob('/'.join((dest, 'Archive-Update-in-Progress-*'))):
        raise ContentBad('Remote archive update in progress.')


@util.run_as_user('www-sync')
def mirror_metadata(archive_ip, dest, link_dest):
    '''Mirror metadata from the archive and verify that it is consistent.
    '''

    # Perform an rsync of the metadata, excluding some contents
    rsync_source = archive_ip + '::ubuntu/'
    rsync_options = ('--quiet',
                     '--recursive',
                     '--links',
                     '--perms',
                     '--chmod=g+w',
                     '--times',
                     '--compress',
                     '--delete',
                     '--delete-during',
                     '--link-dest=%s/' % link_dest,
                     '--exclude=.trace',
                     '--exclude=indices',
                     '--exclude=project/trace',
                     '--exclude=pool',
                     '--exclude=dists/*/*/debian-installer',
                     '--exclude=dists/*/*/dist-upgrader-all',
                     '--exclude=dists/*/*/installer*',
                     '--exclude=dists/*/*/uefi',
                     )
    host.rsync(rsync_source,
               dest,
               options=rsync_options)


@util.run_as_user('www-sync')
def mirror_archive(hostname):
    '''Sync Ubuntu repository metadata to the mirror.

    Metadata synchronization is performed and then the metadata is
    verified.

    Returns metadata version (destination file name)
    '''

    now = datetime.utcnow().replace(microsecond=0)

    apache_data = '/'.join((hookenv.config()['apache-root'], 'data'))
    link_active = '/'.join((apache_data, 'ubuntu_active'))
    link_next = '/'.join((apache_data, 'ubuntu_next'))
    dest = '/'.join((apache_data, 'ubuntu_{}_u{}'.format(
        now.isoformat(sep='_'),
        hookenv.local_unit().split('/')[-1])))
    LOG('Performing sync to %s' % dest)

    # The archive may have multiple IP addresses, try each several times
    archive_ips = [i for i in socket.gethostbyname_ex(hostname)[-1]]

    update_retry = 0
    content_retry = 0
    content_mirrored = False
    for archive_ip in archive_ips:
        LOG('Archive readiness check for %s' % archive_ip)

        need_retry = False
        try:
            mirror_base(archive_ip, dest)
        except CalledProcessError as excp:
            LOG('rsync failed({}): {}'.format(excp.returncode,
                                              excp.output), hookenv.WARNING)
            if util.rsync_should_retry(excp.returncode):
                need_retry = True
            else:
                update_retry = 0
                continue
        except ContentBad as excp:
            LOG('{}'.format(excp.message))
            need_retry = True

        if need_retry:
            update_retry += 1
            if update_retry < 4:
                LOG('Retrying same host. (attempt #%s)' % update_retry)
                archive_ips.insert(0, archive_ip)
                time.sleep(30)
            else:
                LOG('No retries left. (attempt #%s)' % update_retry)
                update_retry = 0
            continue

        LOG('Sync from %s starting.' % archive_ip)
        try:
            mirror_metadata(archive_ip, dest, link_active)
        except CalledProcessError as excp:
            LOG('rsync failed({}): {}'.format(excp.returncode,
                                              excp.output), hookenv.WARNING)
            if util.rsync_should_retry(excp.returncode):
                update_retry += 1
                if update_retry < 4:
                    LOG('Retrying same host. (attempt #%s)' % update_retry)
                    archive_ips.insert(0, archive_ip)
                    time.sleep(30)
                else:
                    LOG('No retries left. (attempt #%s)' % update_retry)
                    update_retry = 0
                continue
            else:
                LOG('rsync failed({}): {}'.format(excp.returncode,
                                                  excp.output),
                    hookenv.WARNING)
                update_retry = 0
                continue
        LOG('Sync from %s completed.' % archive_ip)

        LOG('Starting repository mirror verification.')
        try:
            verify_repository(dest)
        except ContentBad as excp:
            LOG('Verification failed: %s' % excp.message)
            content_retry += 1
            if content_retry < 4:
                LOG('Retrying same host. (attempt #%s)' % update_retry)
                archive_ips.insert(0, archive_ip)
                time.sleep(30)
            else:
                LOG('No retries left. (attempt #%s)' % (update_retry + 1))
                content_retry = 0
            continue
        LOG('Successful repository mirror verification.')
        content_mirrored = True
        break

    if not content_mirrored:
        LOG('Could not gather new metadata from archive.', hookenv.DEBUG)
        current = os.path.realpath(link_active)
        current_good = '_'.join((current, 'good'))
        util.clean_metadata_dir(keep=set((current, current_good)))
        return None

    # Create a canary in the metadata directory to indicate a complete mirror
    open('_'.join((dest, 'good')), 'a').close()

    # link_active will not exist on first run, need to create it
    try:
        os.stat(link_active)
    except OSError, err:
        if err.errno is errno.ENOENT:
            host.symlink(dest, link_active)
        else:
            raise

    # The ubuntu_next link indicates the latest mirrored content
    try:
        os.stat(link_next)
    except OSError, err:
        if err.errno is errno.ENOENT:
            host.symlink(dest, link_next)
        else:
            raise

    # Delete all unused archive metadata mirror directories
    LOG('Removing old, unused repository mirrors.')
    util.clean_metadata_dir(keep=set((os.path.realpath(dest),
                                      os.path.realpath('_'.join((dest,
                                                                 'good'))),
                                      os.path.realpath(link_active))))
    LOG('Unused repository mirrors removed.')

    return os.path.basename(dest)


def parse_release_file(path):
    '''Read a distribution Release file located at 'path'

    Returns a dictionary describing Release file contents.
    '''

    release = {}
    release['name'] = os.path.basename(path.strip('/'))
    release['mtime'] = os.stat(path + '/Release').st_mtime

    # Collect expected SHA256 sum values for files in the distribution
    files = release['files'] = OrderedDict({})
    with open(path + '/Release', 'r') as release_file:
        found = False
        for line in release_file:
            if not found:
                if line.startswith('SHA256:'):
                    found = True
            else:
                if line.startswith(' '):
                    line = line.strip().split()
                    files[line[2]] = {'SHA256': line[0], 'size': line[1]}
                else:
                    break
    return release


def verify_distribution(path, release):
    '''Check distribution file sizes/SHA256sums against Release file values'''

    files = release['files']
    for filename in files:
        filepath = '/'.join((path, filename))

        try:
            disk_size = os.stat(filepath).st_size
        except OSError as excp:
            if excp.errno is errno.ENOENT:
                # There will be a number of files listed in the Release
                # file that are not present on disk.  Alternatively, we
                # could walk release tree and verify against the release
                # metadata
                continue
            else:
                raise

        size = int(files[filename]['size'])
        if disk_size != size:
            msg = 'Sizes do not match for {} in {}'.format(filename,
                                                           release['name'])
            raise ContentBad(msg)

        with open(filepath, 'rb') as infile:
            sha256 = hashlib.sha256()
            for data in infile.read():
                sha256.update(data)
        digest = sha256.hexdigest()
        if digest != files[filename]['SHA256']:
            file_mtime = os.stat(filepath).st_mtime
            msg = 'SHA256 mismatch for {} in {}.'.format(filename,
                                                         release['name'])
            if release['mtime'] < file_mtime:
                ' '.join((msg, 'Stale Release file.'))
            raise ContentBad(msg)


def verify_repository(archive_root):
    '''Verify the content of a Ubuntu software repository's metadata'''

    dists_root = '/'.join((archive_root, 'dists'))

    # Traverse distributions and fail early by starting with the newest.
    # The newest (development branch) will have the highest rate of churn,
    # therefore the greatest chance of being invalid.
    dists = sorted([ent for ent in os.listdir(dists_root)
                    if os.path.isdir('/'.join((dists_root, ent))) and
                    not os.path.islink('/'.join((dists_root, ent)))],
                   reverse=True)

    for dist in dists:
        LOG('Verifying distribution %s' % dist)
        dist_dir = '/'.join((dists_root, dist))
        release = parse_release_file(dist_dir)
        verify_distribution(dist_dir, release)

    for uip in glob.glob('/'.join((archive_root,
                                   'Archive-Update-in-Progress-*'))):
        LOG('Removing {} from clean repository'.format(uip), hookenv.DEBUG)
        os.remove(uip)


def _rsync(source, dest, link_dest=None, optlist=None, user='www-sync',
           retries=1):
    '''Perform an rsync as a specified user'''

    options = ['--quiet',
               '--recursive',
               '--links',
               '--perms',
               '--chmod=g+w',
               '--times',
               '--compress',
               '--delete',
               '--delete-during',
               '--timeout=120',
               '-e', 'sudo -u {} ssh -l {}'.format(user, user),
               ]
    if optlist:
        options.extend(optlist)
    if link_dest:
        options.append('--link-dest=%s/' % link_dest)
    attempt = 0
    while True:
        try:
            host.rsync(source, dest, options=options)
        except CalledProcessError as excp:
            LOG('rsync failed({}): {}'.format(excp.returncode,
                                              excp.output),
                hookenv.WARNING)
            if util.rsync_should_retry(excp.returncode):
                attempt += 1
                if attempt < retries:
                    LOG('Retrying rsync. (attempt #%s)' % attempt)
                    time.sleep(10)
                    continue
                else:
                    LOG('No retries left.')
                    raise
            else:
                LOG('rsync failed({}): {}'.format(excp.returncode,
                                                  excp.output),
                    hookenv.WARNING)
                raise
        return


@util.run_as_user('www-sync')
def rsync_to_peers(source_dir, dest_dir, link_dir=None, user='www-sync',
                   peers=None):
    """
    Use rsync to push a directory to all configured peers.

    Peers are configured in the 'cluster' relationship using
    charmhelpers.contrib.unison.ssh_authorized_peers(); this list can be
    overridden with the peers parameter.

    A list of the peers whose transfers were successful will be
    returned.
    """

    if peers is None:
        peers = unison.collect_authed_hosts(peer_interface='cluster')

    successful_peers = []
    for peer in peers:
        rsync_dest = '{}@{}:{}'.format(user, peer, dest_dir)
        LOG('Syncing {} to {}.'.format(source_dir, rsync_dest))
        try:
            _rsync(source_dir, rsync_dest, link_dest=link_dir,
                   user=user, retries=10)
        except Exception:
            LOG('Sync to {} failed'.format(peer))
            traceback.print_exc(file=sys.stdout)
        else:
            successful_peers.append(peer)
            LOG('Sync to {} complete.'.format(peer))
    return successful_peers


@util.run_as_user('www-sync')
def rsync_from_peer(source_dir, dest_dir, peer_ip, link_dir=None,
                    user='www-sync'):
    '''Use rsync to pull a directory from a configured peer.  '''

    rsync_source = '{}@{}:{}'.format(user, peer_ip, source_dir)
    LOG('Syncing {} to {}.'.format(rsync_source, dest_dir))
    _rsync(rsync_source, dest_dir, link_dest=link_dir,
           user=user, retries=10)
    LOG('Sync to {} complete.'.format(peer_ip))


def _leader_update_metadata():
    '''Update the metadata on a leader in the cluster.

    If the leader is newly elected, it will pull down the latest
    metadata.  The metadata is not pushed to peers.  The peers will
    pull the metadata once on the next relation-changed hook triggered
    by setting the meta_version relation data.
    '''

    LOG('Updating metadata on the leader')
    # Relation data
    rel = hookenv.relations()[hookenv.relation_type()][hookenv.relation_id()]
    leader_id = sorted([hookenv.local_unit()] + cluster.peer_units())[0]
    leader_rel = rel[leader_id]

    active_ver = util.get_active_metaver()
    meta_ver = leader_rel.get('meta_version')
    apache_data = '/'.join((hookenv.config()['apache-root'], 'data'))
    if meta_ver:
        meta_dir = '/'.join((apache_data, meta_ver))
        canary = '_'.join((meta_dir, 'good'))

    # Common case.  All good, just need to update to latest metadata
    if meta_ver and os.path.exists(meta_dir) and os.path.exists(canary):
        if meta_ver != active_ver:
            LOG('Updated to metadata version {}'.format(meta_ver))
            util.update_active_symlink(meta_ver)
        else:
            LOG('Nothing to update')
    else:
        LOG('Leader missing expected metadata. Recovering.')
        LOG('meta_ver:{} active_ver:{}'.format(meta_ver, active_ver),
            hookenv.DEBUG)
        util.clean_metadata_dir()

        # Don't use the existing metadata.  If this is a current unit
        # taking over from a prior leader, use of the existing metadata
        # could move units back in time.
        new_ver = mirror_archive(hookenv.config()['sync-host'])
        if not new_ver:
            LOG('Leader could not gather new metadata', hookenv.WARNING)
            return
        util.update_active_symlink(new_ver)

        LOG('Updating metadata version to {}'.format(new_ver),
            hookenv.DEBUG)
        hookenv.relation_set(meta_version=new_ver)

    service.unpause()
    util.clean_metadata_dir()
    LOG('Metadata update complete.')


def _nonleader_update_metadata():
    '''Update the metadata on a non-leader peer in the cluster'''

    LOG('Updating metadata on a peer')

    rel = hookenv.relations()[hookenv.relation_type()][hookenv.relation_id()]
    local_rel = rel[hookenv.local_unit()]
    leader_id = sorted([hookenv.local_unit()] + cluster.peer_units())[0]
    leader_rel = rel[leader_id]

    active_ver = util.get_active_metaver()
    meta_ver = leader_rel.get('meta_version')
    apache_data = '/'.join((hookenv.config()['apache-root'], 'data'))
    if meta_ver:
        meta_dir = '/'.join((apache_data, meta_ver))
        canary = '_'.join((meta_dir, 'good'))

    # Clean up meta_version on local relation if no longer the leader
    if local_rel.get('meta_version'):
        LOG('Clearing metadata', hookenv.DEBUG)
        hookenv.relation_set(meta_version=None)

    if not leader_rel.get('ssh_pub_key'):
        LOG('SSH keys not yet present for leader.', hookenv.INFO)
        return
    if not meta_ver:
        LOG('Metadata version not yet provided by leader.', hookenv.INFO)
        return

    # Need to update metadata
    if meta_ver and meta_ver != active_ver:
        # Already have good metadata
        if os.path.exists(meta_dir) and os.path.exists(canary):
            LOG('Updated to metadata version {}'.format(meta_ver))
            util.update_active_symlink(meta_ver)
            if util.in_failover():
                service.unpause()

        # Need to pull metadata from leader
        else:
            # Pause service and make space for new metadata
            service.pause()
            util.clean_metadata_dir()

            # Can we connect to the leader?
            leader_hosts = leader_rel.get('ssh_authorized_hosts')
            if not leader_hosts or \
                    hookenv.unit_private_ip() not in leader_hosts:
                LOG('Leader has not added IP to authorized_hosts yet, '
                    'can not update.')
                return

            # Pull metadata from the leader if it exists and is ready
            leader_ip = leader_rel.get('private-address')
            if util.remote_path_exists('www-sync', leader_ip, canary):
                LOG('Pulling metadata {} from leader.'.format(meta_ver))
                active_dir = '/'.join((apache_data, 'ubuntu_active'))
                source = meta_dir + '/'
                dest = meta_dir
                try:
                    rsync_from_peer(source, dest,
                                    leader_rel.get('private-address'),
                                    link_dir=active_dir)
                except CalledProcessError:
                    traceback.print_exc()
                    LOG('Failed to pull metadata from leader; service stopped'
                        ' until next update from leader.')
                else:
                    util.update_active_symlink(meta_ver)
                    service.unpause()

            # Do nothing. Service is stopped until next update from leader
            else:
                LOG('Metadata dir does not exist on leader.',
                    hookenv.DEBUG)
    util.clean_metadata_dir()
    LOG('Metadata update complete.')


def peer_update_metadata():
    '''Update the metadata on a clustered peer to the current version.

    If the 'cluster' relationship exists, this will look for the
    current 'meta_version' key on the leader and attempt to
    update the active metadata to this version.

    This function should be called from the cluster-relation-joined,
    cluster-relation-changed, and cluster-relation-departed hooks.
    '''

    if not hookenv.is_relation_made('cluster') or \
            not hookenv.in_relation_hook() or \
            not hookenv.relation_type() == 'cluster':
        return

    if util.is_elected_leader(None):
        # Leader does not pause, a newly appointed leader must be available
        if util.in_failover():
            service.unpause()
        _leader_update_metadata()
    else:
        _nonleader_update_metadata()


def update_checks(nrpe_config):
    '''Update nagios check for mirror metadata age'''

    service_script = '/'.join((hookenv.charm_dir(),
                               'files',
                               'nrpe-external-master',
                               'check_metadata_age'))
    metadata_dir = '/'.join((hookenv.config()['apache-root'],
                             'data',
                             'ubuntu_active'))

    nrpe_config.add_check(
        shortname='mirror',
        description='Ubuntu repository cache metadata age',
        check_cmd=' '.join((service_script, metadata_dir)),
        )
