'''Functions related to squid-deb-proxy use within this charm'''

import os
import subprocess

# pylint can't find the modules  # pylint: disable=F0401
from charmhelpers import fetch

from charmhelpers.core import (
    hookenv,
    host,
    templating,
)

import util
# pylint: enable=F0401

LOG = hookenv.log
SERVICE = 'squid-deb-proxy'


@util.run_as_user('proxy')
def init_caches():
    '''Initialize the squid on-disk caches

    This must be called while the service is stopped and after the
    configuration has been written.
    '''

    if not hookenv.config()['squid-disk-caches']:
        LOG('No on disk cache directories founds for squid.',
            hookenv.ERROR)
        return
    subprocess.check_call(['/usr/sbin/squid3', '-N', '-z',
                           '-f', '/etc/squid-deb-proxy/squid-deb-proxy.conf'])


def size_caches():
    '''Determine the sizes for in-memory/on-disk squid caches

    This must be run at installation time.  Runtime tuning is unsupported.
    '''

    config = hookenv.config()
    avail_disk = 0
    for cache in config['squid-disk-caches']:
        avail_disk += cache[1]
    LOG('{}MB of disk cache available for squid.'.format(avail_disk),
        hookenv.DEBUG)

    memory = util.system_total_mem() / 2
    # Estimate of 100MB of overhead for squid
    memory = max(memory - 100, 0)

    if memory < 128:
        LOG('Squid has too little memory ({}MB), '
            'can not deploy service'.format(memory),
            hookenv.ERROR)
        raise RuntimeError('Too little system memory for squid deployment')

    # Start by giving squid a maximum of 256MB for in-memory cache
    config['squid-cache-mem'] = min(memory, 256)
    memory -= config['squid-cache-mem']

    # Estimate 20MB of memory overhead per GB of on-disk cache
    max_disk = int(memory / 20.0 * 1024)
    config['squid-cache-disk'] = min(max_disk, avail_disk)
    disk_overhead = int(round(config['squid-cache-disk'] / 1024.0 * 20))
    memory -= disk_overhead

    # Remaining memory (up to 1/2 total RAM) is given to in-memory cache
    config['squid-cache-mem'] += memory

    LOG('Squid sizing complete. In-memory cache: {}MB '
        'On-disk cache: {}MB'.format(config['squid-cache-mem'],
                                     config['squid-cache-disk']),
        hookenv.INFO)

    config.save()


def install():
    '''Perform the install/config and disable avahi'''

    LOG('Installing squid', hookenv.INFO)
    fetch.apt_install(SERVICE, fatal=True)

    # We don't need avahi-daemon or squid3 running
    host.service_stop('avahi-daemon')
    with open('/etc/init/avahi-daemon.override', 'w') as avahi_override_file:
        avahi_override_file.write('manual\n')

    host.service_stop('squid3')
    with open('/etc/init/squid3.override', 'w') as squid3_override_file:
        squid3_override_file.write('manual\n')

    stop()

    size_caches()
    render_configs()
    init_caches()


@util.run_as_user('root')
def start():
    '''Start the service and remove the override file blocking startup'''

    LOG('Starting squid', hookenv.INFO)
    override_path = '/etc/init/{}.override'.format(SERVICE)
    if os.path.exists(override_path):
        os.remove(override_path)

    host.service_start(SERVICE)


@util.run_as_user('root')
def stop():
    '''Stop the service and block startup at boot'''

    LOG('Stopping squid', hookenv.INFO)
    with open('/etc/init/{}.override'.format(SERVICE), 'w+') as override_file:
        override_file.write('manual\n')
    host.service_stop(SERVICE)


@util.run_as_user('root')
def service_running():
    '''Return True is the service is running'''

    return host.service_running(SERVICE)


def peer_config():
    '''Return the configuration strings for squid cache peers'''

    config = ''
    for ip in util.peer_IPs():
        config = '\n'.join((config, 'cache_peer %s sibling 8000 3130' % ip))
    return config


def cache_dirs():
    '''Generate 'cache_dir' lines for the squid configuration file.

    Returns a string containing the cache_dir line(s)
    '''
    config = hookenv.config()
    cache_str = ''
    max_size = config['squid-cache-disk']
    for path, free in config['squid-disk-caches']:
        size = min(max_size, free)
        max_size -= size
        cache_str = ''.join((cache_str,
                             'cache_dir aufs {} {} 16 256\n'.format(path,
                                                                    size)))
    return cache_str


@util.run_as_user('root')
def render_configs():
    '''Render the configuration templates for all required services'''

    LOG('Rendering squid configuration templates')
    config = hookenv.config()
    context = {}
    context['Cache_Config'] = cache_dirs()
    context['Cache_Mem'] = config['squid-cache-mem']
    context['Cache_Peers'] = peer_config()
    context['Cache_Hostname'] = hookenv.unit_private_ip()
    templating.render('{}/{}.conf'.format(SERVICE, SERVICE),
                      '/etc/{}/{}.conf'.format(SERVICE, SERVICE),
                      context)

    peer_context = {}
    peer_context['Cache_Sibling_IPs'] = '\n'.join(util.peer_IPs())
    templating.render('{}/allowed-networks-src.conf'.format(SERVICE),
                      '/etc/{}/allowed-networks-src.acl'.format(SERVICE),
                      peer_context)


def update_checks(nrpe_config):
    ''''Update Nagios checks for squid serving archive pool content'''

    nrpe_config.add_check(
        shortname='squid',
        description='squid serving archive pool',
        check_cmd='check_http --hostname localhost --url "/ubuntu/pool/" '
                  '--string "main/" --expect \"200\\ OK\"',
        )
